import numpy as np
import torch
import torch.nn as nn

from linear import *

dim_i=100
dim_o=50
N=1000
model = C_Linear_Binary_Sigmoid(dim_i,dim_o)
L = 1.0*(torch.randn(dim_i,dim_o)-1.5>0)
# model = nn.Linear(100,50,dtype=torch.complex64)
# x = torch.randn(20,100,dtype=torch.complex64)
# y = torch.randn(20,50,dtype=torch.complex64)
opti = torch.optim.Adam(model.parameters(),lr=1e-2)
for i in range(1000):
	x = torch.randn(N,dim_i)
	y = torch.einsum("ij,ni->nj",L,x)
	M = model(x)
	err = M-y
	err = torch.mean(torch.conj(err)*err).real
	opti.zero_grad()
	err.backward()
	opti.step()
	print(f"{i:05d} | {err.item()}")

import matplotlib.pyplot as plt
plt.ion()
truth=L
fnd=model.L.detach()
found=1.0*(model.L>0)
# plt.imshow(L)
# plt.imshow(truth-found + 2*truth)
fig, axs = plt.subplots(1,3)
axs[0].imshow(L)
axs[1].imshow(truth-found + 2*truth)
axs[2].imshow(torch.stack((
	(fnd>0)*fnd,
	(fnd<0)*1.,
	(fnd<0)*-fnd
),2))
plt.ioff()

####################################################################################################
####################################################################################################
####################################################################################################

from toy_problem_mnist_analysis import analysis_classification
from toy_problem_mnist import mnist_fc

# model = nn.Sequential(*mnist_fc(layers,σ=nn.GELU(),linear_gen=lambda λ1,λ2: C_Linear_Binary_Sigmoid(λ1,λ2)))
# identifier = "C_Linear_Binary_Sigmoid(5)"
# analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=45,batch_size=256,fashion=True)

def debug(lin,name,layer_N = 5):
	layers = np.exp(np.linspace(np.log(28*28),np.log(10),layer_N))
	model = nn.Sequential(*mnist_fc(layers,σ=nn.GELU(),linear_gen=lambda λ1,λ2: lin(λ1,λ2)))
	identifier = f"{name}({layer_N})"
	analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=45,batch_size=256,fashion=True)

# debug(Linear_Funky,"Linear_Funky",layer_N = 5)
# debug(C_Linear,"C_Linear",layer_N = 5)
# debug(C_Linear_2_Softmax,"C_Linear_2_Softmax",layer_N = 5)
# debug(C_Linear_Binary_Sigmoid,"C_Linear_Binary",layer_N = 5)
# debug(C_Linear_Sigmoid,"C_Linear_Binary_Sigmoid",layer_N = 5)
# debug(C_Linear_Ternary,"C_Linear_Ternary",layer_N = 5)
# debug(C_Linear_Sigmoid_Ternary,"C_Linear_Ternary_Sigmoid",layer_N = 5)

####################################################################################################
# FUNKY VS NORMAL ##################################################################################
####################################################################################################

layer_N = 5
layers = np.exp(np.linspace(np.log(28*28),np.log(10),layer_N))

# # model = nn.Sequential(*mnist_fc(layers,σ=nn.GELU(),linear_gen=lambda λ1,λ2: Linear_Funky(λ1,λ2)))
# # identifier = "funky(5)"
# # analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=128,batch_size=64,fashion=True)

# # model = nn.Sequential(*mnist_fc(layers,σ=nn.GELU(),linear_gen=lambda λ1,λ2: nn.Linear(λ1,λ2)))
# # identifier = "normal(5)"
# # analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=128,batch_size=64,fashion=True)

# # funky = h5py.File(f'out1/funky(5).h5', 'r')
# # normal = h5py.File(f'out1/normal(5).h5', 'r')
# # plt.plot(np.array(funky["perf_T_accuracy"]))
# # plt.plot(np.array(normal["perf_T_accuracy"]))

# # model = nn.Sequential(*mnist_fc(layers,σ=nn.GELU(),linear_gen=lambda λ1,λ2: nn.Linear(λ1,λ2)))
# # identifier = "normal(5)"
# # analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=45,batch_size=256,fashion=True)

# for i in range(20):
# 	bs = 256
# 	epochs=45
# 	model = nn.Sequential(*mnist_fc(layers,σ=nn.GELU(),linear_gen=lambda λ1,λ2: Linear_Funky(λ1,λ2)))
# 	identifier = f"funky(5)_{i}"
# 	analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=epochs,batch_size=bs,fashion=True)
# 	#
# 	model = nn.Sequential(*mnist_fc(layers,σ=nn.GELU(),linear_gen=lambda λ1,λ2: nn.Linear(λ1,λ2)))
# 	identifier = f"normal(5)_{i}"
# 	analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=epochs,batch_size=bs,fashion=True)

# import h5py
# import matplotlib.pyplot as plt
# plt.ion()

# for i in range(20):
# 	identifier = f"funky(5)_{i}"
# 	funky = h5py.File(f'out/{identifier}.h5', 'r')
# 	identifier = f"normal(5)_{i}"
# 	normal = h5py.File(f'out/{identifier}.h5', 'r')
# 	plt.plot(np.array(funky["perf_T_accuracy"]),color="orange")
# 	plt.plot(np.array(normal["perf_T_accuracy"]),color="blue")
